import torch
from torch import nn
import numpy as np 
from torch.cuda.amp import  autocast
from torch.nn import functional as F
from typing import List, Optional, Tuple, Union

## -- model -- ##


def init_layer(layer):
    """Initialize a Linear or Convolutional layer. """
    nn.init.xavier_uniform_(layer.weight)
 
    if hasattr(layer, 'bias'):
        if layer.bias is not None:
            layer.bias.data.fill_(0.)
            
    
def init_bn(bn):
    """Initialize a Batchnorm layer. """
    bn.bias.data.fill_(0.)
    bn.weight.data.fill_(1.)
    
class AttBlock(nn.Module):
    def __init__(self, n_in, n_out, activation='linear', temperature=1.):
        super(AttBlock, self).__init__()
    
        self.activation = activation
        self.temperature = temperature
        self.att = nn.Conv1d(in_channels=n_in, out_channels=n_out, kernel_size=1, stride=1, padding=0, bias=True)
        self.cla = nn.Conv1d(in_channels=n_in, out_channels=n_out, kernel_size=1, stride=1, padding=0, bias=True)
        
        #self.bn_att = nn.BatchNorm1d(n_out)
        self.init_weights()
        
    def init_weights(self):
        init_layer(self.att)
        init_layer(self.cla)
        #init_bn(self.bn_att)
         
    def forward(self, x): # (bs, hidden size, step/channel)
        att =   torch.clamp(self.att(x), -10, 10) #  torch.tanh(self.att(x)) #
        norm_att = torch.softmax(att, dim=-1) # apply softmax on step dim
        cla = self.nonlinear_transform(self.cla(x)) 
        x = torch.sum(norm_att * cla, dim=2)
        return x, norm_att, cla

    def nonlinear_transform(self, x):
        if self.activation == 'linear':
            return x
        elif self.activation == 'sigmoid':
            return torch.sigmoid(x)
        elif self.activation == 'softmax':
            return torch.softmax(x, dim=-1)
        else:
            raise("Attention Block activation function not accepted. Use only [linear, sigmoid, softmax]")
        
class GlobalAttention(nn.Module):
    def __init__(self, hidden_size):
        super(GlobalAttention , self).__init__()
        self.W = nn.Linear(hidden_size, hidden_size, bias=False)

    def forward(self, h_i):
        u_i = self.W(h_i)
        a_i =  F.softmax(u_i, dim=1) 
        v = (a_i * h_i).sum(1)   
        return v, a_i

class Transpose(nn.Module):
    def __init__(self, axis1, axis2):
        super(Transpose , self).__init__()
        self.axis1 = axis1 
        self.axis2 = axis2

    def forward(self, inputs):
        return torch.transpose(inputs, self.axis1, self.axis2)

class SED_Block(nn.Module):
    def __init__(self, num_classes:List[int], in_features:int, hidden_size:int = 1024, activation:int = "sigmoid",
                drop_rate:List[int] = [0.5, 0.5], apply_attention="step"):
        """Apply SED block on output of a backbone.
        """
        super(SED_Block, self).__init__()
        assert apply_attention in ["step", "channel"]
        self.apply_attention = apply_attention
        self.num_classes = num_classes
        #self.dropout = nn.Dropout(drop_rate[0])
        self.blocks = nn.ModuleList()
        for num_class in self.num_classes:
            # 1 ) att on step => (bs, channel, step) = > transpose => (bs, step, channel) = > linear => (bs, step, hidden size)
            # => transpose => (bs, hidden size, step) => attblock on hstep => (bs , num_classes, step )
            # 2 ) att on channels => (bs, channel, step) = > linear => (bs, channel, hidden size)
            # => transpose => (bs, hidden size, channel) => attblock on channel => (bs , num_classes, channel  ) => (bs , num_classes,   )
            self.blocks.append(nn.Sequential(nn.Dropout(drop_rate[0]),
                                            Transpose(1,2) if self.apply_attention == "step" else nn.Identity(), 
                                            nn.Linear(in_features, hidden_size),
                                            nn.ReLU(),
                                            Transpose(1,2) ,
                                            nn.Dropout(drop_rate[1]),
                                            AttBlock(hidden_size,  num_class, activation=activation, temperature=1.)))

    def forward(self, inputs):
        """
            inputs: (Batch size, channels, steps, freq)
            @return: List[num_classes] each element of the list following this format : 
                    {'clipwise':(BATCH SIZE, n), 'segmentwise':(BATCH SIZE, step, n),
                    'norm_att':(BATCH SIZE, step)}
        """
        outputs = []
        x = torch.mean(inputs, dim=3) # Average freq features
    
        # smooth channels
        x1 = F.max_pool1d(x, kernel_size=3, stride=1, padding=1) # pool on steps
        x2 = F.avg_pool1d(x, kernel_size=3, stride=1, padding=1)
        x = x1 + x2

        for block in self.blocks:
            (clipwise_output, norm_att, segmentwise_output) = block(x)
            segmentwise_output = segmentwise_output.transpose(1, 2) # batch size, steps, channels
            outputs.append({"clipwise":clipwise_output, "segmentwise":segmentwise_output, 
            "norm_att":norm_att})
        return outputs
        





class ConfigSED(object):
    def __init__(self, window='hann', center=True, pad_mode='reflect', windows_size=2048, hop_size=512,
                sample_rate=32000, mel_bins=128, fmin=50, fmax=16000, ref=1.0, amin=1e-10, top_db=None):
        # Spectrogram
        self.window = window
        self.center = center 
        self.pad_mode = pad_mode
        self.windows_size = windows_size 
        self.hop_size = hop_size

        # Logmel Spectrogram
        self.sample_rate=sample_rate
        self.mel_bins=mel_bins 
        self.fmin=fmin 
        self.fmax=fmax
        self.ref = ref 
        self.amin = amin 
        self.top_db = top_db


from tensorflow.keras import initializers, Sequential
from tensorflow.keras.layers import Dropout, Permute, Dense,   AveragePooling1D, MaxPooling1D
import tensorflow as tf 
from typing import List, Optional, Tuple, Union

class AttBlock(tf.keras.layers.Layer):
    def __init__(self, n_in, n_out, activation='linear', temperature=1.,**kwargs):
        super(AttBlock, self).__init__(**kwargs)
        
        self.activation = activation
        self.temperature = temperature
        self.att = tf.keras.layers.Conv1D(filters=n_out, kernel_size=1, strides=1, padding="valid", use_bias=True, name="att", #activation="tanh",
                                          bias_initializer=initializers.Zeros(), kernel_initializer=tf.initializers.GlorotUniform())
        self.cla = tf.keras.layers.Conv1D(filters=n_out, kernel_size=1, strides=1, padding="valid", use_bias=True, activation=activation, name="cla", 
                                          bias_initializer=initializers.Zeros(), kernel_initializer=tf.initializers.GlorotUniform())
        
        #self.bn_att = tf.keras.layers.BatchNormalization(name="bn_att", beta_initializer=initializers.Zeros(), gamma_initializer=initializers.Ones())#(n_out)

         
    def call(self, x, training=False):
        # x: (n_samples, step, channel)
        att =   tf.clip_by_value(self.att(x), -10, 10)  
        norm_att = tf.nn.softmax(att, axis=1) # apply softmax on step dim
        cla = self.cla(x)
        x = tf.math.reduce_sum(norm_att * cla, axis=1)
        return x, norm_att, cla

        
class GlobalAttention(tf.keras.layers.Layer):
    def __init__(self, hidden_size, **kwargs):
        super(GlobalAttention , self).__init__(**kwargs)
        self.W = tf.keras.layers.Dense(hidden_size, use_bias=False, name="W")

    def call(self, h_i, training=False): # (bs, step, proba)
        u_i = self.W(h_i) # (bs, step, proba)
        a_i =  tf.nn.softmax(u_i, axis=1)#µ/2.0
        v = tf.math.reduce_sum( a_i * h_i, axis=1) 
        return v, a_i

class Identity(tf.keras.layers.Layer):
    def __init__(self ):
        super(Identity , self).__init__()
        self.act = tf.keras.layers.Activation("linear")
    def call(self, inputs):
        return self.act(inputs)

class SED_Block(tf.keras.layers.Layer):
    def __init__(self, num_classes:List[int], in_features:int, hidden_size:int = 1024, activation:int = "sigmoid",
                drop_rate:List[int] = [0.5, 0.5], apply_attention="step"):
        """Apply SED block on output of a backbone.
        """
        super(SED_Block, self).__init__()
        assert apply_attention in ["step", "channel"]
        self.apply_attention = apply_attention
        self.num_classes = num_classes
        #self.dropout = nn.Dropout(drop_rate[0])
        self.max_pool_1d = MaxPooling1D(pool_size=3, strides=1, padding="same")
        self.avg_pool_1d = AveragePooling1D(pool_size=3, strides=1, padding="same")
        self.blocks = []
        self.attn_blocks = []
        for num_class in self.num_classes:
            # 1 ) att on step => (bs, channel, step) = > transpose => (bs, step, channel) = > linear => (bs, step, hidden size)
            # => transpose => (bs, hidden size, step) => attblock on hstep => (bs , num_classes, step )
            # 2 ) att on channels => (bs, channel, step) = > linear => (bs, channel, hidden size)
            # => transpose => (bs, hidden size, channel) => attblock on channel => (bs , num_classes, channel  ) => (bs , num_classes,   )
            self.blocks.append(Sequential([Dropout(drop_rate[0]),
                                             Permute((1,2)) if self.apply_attention == "step" else Identity(), 
                                            Dense(hidden_size, activation="relu"),
                                            Permute((1,2)) ,
                                            Dropout(drop_rate[1])]))
            self.attn_blocks.append(AttBlock(hidden_size,  num_class, activation=activation, temperature=1.))

    def call(self, inputs, return_dict=False):
        """
            inputs: (Batch size, steps, freq, channels)
            @return: List[num_classes] each element of the list following this format : 
                    {'clipwise':(BATCH SIZE, n), 'segmentwise':(BATCH SIZE, step, n),
                    'norm_att':(BATCH SIZE, step)}
        """
        outputs = []
        x = tf.reduce_mean(inputs, axis=2) # Average freq features
    
        # smooth channels
        x1 = self.max_pool_1d(x) # pool on steps
        x2 = self.avg_pool_1d(x)
        x = x1 + x2

        for attn_block, block in zip(self.attn_blocks, self.blocks):
            o = block(x)
            (clipwise_output, norm_att, segmentwise_output) = attn_block(o)
            segmentwise_output = tf.transpose(segmentwise_output, perm=[0,2,1]) # batch size, steps, channels
            if return_dict:
                outputs.append({"clipwise":clipwise_output, "segmentwise":segmentwise_output, "norm_att":norm_att})
            else:
                outputs.append(clipwise_output)
        return outputs


from audio_sed.pytorch.sed_block import SED_Block 
from audio_sed.pytorch.sed_models import AudioSED,  shape_from_backbone, AudioClassifierSequenceSED
import torch 
from torch import nn 
import numpy as np 
import timm
from audio_sed.sed_config import ConfigSED
import pytest


@pytest.mark.unittest
def test_sed_block_pt():
    # (Batch size, channels, steps, freq)

    batch_size = 4
    channels = 16
    steps = 12 
    freq = 64
    num_classes = [10, 2]
    inputs = torch.as_tensor(np.random.uniform(0, 1, (batch_size, channels, steps, freq))).float()
    block = SED_Block(num_classes=num_classes, in_features=channels, hidden_size=1024, activation="sigmoid",
                    drop_rate=[0.5, 0.5], apply_attention="step")
    with torch.no_grad():
        output = block(inputs)
    assert len(output) == len(num_classes)
    assert type(output[0]) == dict

@pytest.mark.unittest
def test_model_pt():
    batch_size = 4
    channels = 16
    steps = 12 
    freq = 64
    num_classes = [10, 2]
    num_step = 3
    inputs = torch.as_tensor(np.random.uniform(0, 1, (batch_size, channels, steps, freq))).float()

    backbone = timm.create_model('tf_efficientnet_b0.ns_jft_in1k', num_classes=0)
    backbone.global_pool = nn.Identity()

    inputs = torch.as_tensor(np.random.uniform(0, 1, (batch_size, 32000*5))).float()
    inputs2 = torch.as_tensor(np.random.uniform(0, 1, (batch_size, num_step, 32000*5))).float()

    shape =  shape_from_backbone(inputs, backbone, use_logmel=True)
    model = AudioSED( backbone, num_classes=num_classes, in_features=shape[1], hidden_size=1024, activation= 'sigmoid', use_logmel=True, 
                    spectrogram_augmentation = None, apply_attention="step")

    model_seq = AudioClassifierSequenceSED(model)
    with torch.no_grad():
        output = model(inputs)
        output2 = model_seq(inputs2)

    assert type(output) == list
    assert len(output) == len(num_classes)
    assert output[0]['clipwise'].shape == (batch_size, num_classes[0]) and output[1]['clipwise'].shape == (batch_size, num_classes[1])

    assert type(output2) == list
    assert len(output2) == len(num_classes)
    assert output2[0]['output'].shape == (batch_size, num_classes[0]) and output2[1]['output'].shape == (batch_size, num_classes[1])
    assert output2[0]['clipwise'].shape == (batch_size, num_step, num_classes[0]) and output2[1]['clipwise'].shape == (batch_size, num_step, num_classes[1])

if __name__ == "__main__":
    pass


# -*- coding: utf-8 -*-
from setuptools import setup
from io import open

def readme():
    with open('README.md', encoding="utf-8-sig") as f:
        README = f.read()
    return README


setup(
    name='audio_sed',
    version='0.0.2',    
    description='Implementation of audio SED architecture for tensorflow and pytorch.',
    author='Shiro-LK',
    author_email='shirosaki94@gmail.com',
    license='MIT License',
    packages=['audio_sed', 'audio_sed.pytorch', 'audio_sed.tensorflow'],
    long_description=readme(),
    long_description_content_type="text/markdown",
    install_requires=['numpy', 
                      "pytest", "tensorflow", "torch",
                      ],
    url='https://github.com/Shiro-LK/SoundEventDetection',
    download_url='https://github.com/Shiro-LK/SoundEventDetection.git',
    keywords=["torch", "tensorflow", "audio_sed"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
)



import numpy as np 
import random
import math

class PermutationChunks(object):
    def __init__(self, min_, max_, sr=32000):
        self.min_ = min_
        self.max_ = max_
        self.sr = sr 
    
    def __call__(self, samples):
        resize = False 
        shape = samples.shape
        if len(shape) == 2:
            samples = samples.reshape(-1,)
            resize = True
        N = int(math.floor(len(samples)/self.sr) )

        vals = np.random.uniform(self.min_, self.max_, (N,))
        chunks = []
        i = 0
        for v in vals:
            t = int(v * self.sr)
            chunks.append(samples[i:i+t])
            i = i+t
            if i > len(samples):
                break
        random.shuffle(chunks)
        chunks = np.concatenate(chunks)
        if resize:
            chunks = chunks.reshape(*shape)
        return chunks 


from audio_sed.tensorflow.sed_block import SED_Block 
from audio_sed.tensorflow.sed_models import AudioSED,  shape_from_backbone, AudioClassifierSequenceSED
import tensorflow as tf
import numpy as np 
import tfimm
from audio_sed.sed_config import ConfigSED
import pytest

@pytest.mark.unittest
def test_sed_block():
    batch_size = 4
    channels = 16
    steps = 12 
    freq = 64
    num_classes = [10, 2]
    
    inputs = np.random.uniform(0, 1, (batch_size, steps, freq, channels))

    block = SED_Block(num_classes=num_classes, in_features=channels, hidden_size=1024, activation="sigmoid",
                    drop_rate=[0.5, 0.5], apply_attention="step")
    output = block(inputs, return_dict=True)
    assert len(output) == len(num_classes)
    assert type(output[0]) == dict

    output = block(inputs, return_dict=False)
    assert len(output) == len(num_classes)

@pytest.mark.unittest 
def test_model():
    batch_size = 4
    channels = 16
    steps = 12 
    freq = 64
    num_step = 3
    num_classes = [10, 2]
    
    inputs = np.random.uniform(0, 1, (batch_size, steps, freq, channels))
    backbone = tfimm.create_model('efficientnet_b0_ns', nb_classes=0)
    backbone.pool = tf.keras.layers.Activation("linear")
    backbone.flatten = tf.keras.layers.Activation("linear")
    backbone.dropout = tf.keras.layers.Activation("linear")

    inputs =  np.random.uniform(0, 1, (batch_size, 32000*5))  
    inputs2 =  np.random.uniform(0, 1, (batch_size, num_step, 32000*5)) 

    shape =  shape_from_backbone(inputs, backbone, use_logmel=True)
    model = AudioSED( backbone, num_classes=num_classes, in_features=shape[1], hidden_size=1024, activation= 'sigmoid', use_logmel=True, 
                    spectrogram_augmentation = None, apply_attention="step")

    model_seq = AudioClassifierSequenceSED(model)
    
    output = model(inputs)
    assert type(output) == list
    assert len(output) == len(num_classes)
    assert output[0].shape == (batch_size, num_classes[0]) and output[1].shape == (batch_size, num_classes[1])

    output2 = model_seq(inputs2, return_dict=True)
    print(output2[0]['output'])
    assert type(output2) == list
    assert len(output2) == len(num_classes)
    assert output2[0]['output'].shape == (batch_size, num_classes[0]) and output2[1]['output'].shape == (batch_size, num_classes[1])
    assert output2[0]['clipwise'].shape == (batch_size, num_step, num_classes[0]) and output2[1]['clipwise'].shape == (batch_size, num_step, num_classes[1])



    output2 = model_seq(inputs2, return_dict=False)
    assert type(output2) == list
    assert len(output2) == len(num_classes)
    assert output2[0].shape == (batch_size, num_classes[0]) and output2[1].shape == (batch_size, num_classes[1])


if __name__ == "__main__":
    pass


from tflibrosa.stft import Spectrogram, LogmelFilterBank
from audio_sed.sed_config import ConfigSED
from audio_sed.tensorflow.sed_block import SED_Block, GlobalAttention
import tensorflow as tf 
from tensorflow.keras.layers import Dense, BatchNormalization

def shape_from_backbone(inputs, model, use_logmel=True, config_sed = ConfigSED().__dict__ ):
    sample_rate = config_sed["sample_rate"]
    window_size = config_sed["windows_size"]
    hop_size = config_sed["hop_size"]
    window = config_sed["window"]
    center = config_sed["center"]
    pad_mode = config_sed["pad_mode"]
    mel_bins = config_sed["mel_bins"]
    fmin = config_sed["fmin"]
    fmax = config_sed["fmax"]
    ref = config_sed["ref"]
    amin = config_sed["amin"]
    top_db = config_sed["top_db"]

    spectrogram_extractor = Spectrogram(n_fft=window_size, hop_length=hop_size, 
            win_length=window_size, window=window, center=center, pad_mode=pad_mode, 
            freeze_parameters=True)

    # Logmel feature extractor
    logmel_extractor = LogmelFilterBank(sr=sample_rate, n_fft=window_size, 
        n_mels=mel_bins, fmin=fmin, fmax=fmax, ref=ref, amin=amin, 
        top_db=top_db, freeze_parameters=True)

     
    x =  spectrogram_extractor(inputs)   # (batch_size, 1, time_steps, freq_bins)
    print("sepctrogram :", x.shape)
    if use_logmel:
        x =  logmel_extractor(x) # (batch_size, 1, time_steps, mel_bins)
        print("logmel:", x.shape)
    x = model(tf.concat([x, x, x], axis=-1))
    return x.shape


class AudioSED(tf.keras.Model):
    def __init__(self, backbone, num_classes:list, in_features:int, hidden_size=1024, activation= 'sigmoid', use_logmel=True, 
                spectrogram_augmentation = None, apply_attention="step", drop_rate = [0.5, 0.5], config_sed:dict = ConfigSED().__dict__):
        """
            Classifier for a new task using pretrained CNN as a sub module.
        """
        super(AudioSED, self).__init__()    
        self.num_classes = num_classes
        self.use_logmel = use_logmel
        self.sample_rate = config_sed["sample_rate"]
        self.window_size = config_sed["windows_size"]
        self.hop_size = config_sed["hop_size"]
        self.window = config_sed["window"]
        self.center = config_sed["center"]
        self.pad_mode = config_sed["pad_mode"]
        self.mel_bins = config_sed["mel_bins"]
        self.fmin = config_sed["fmin"]
        self.fmax = config_sed["fmax"]
        self.ref = config_sed["ref"]
        self.amin = config_sed["amin"]
        self.top_db = config_sed["top_db"]
        self.spectrogram_augmentation = spectrogram_augmentation
        self.bn = BatchNormalization()
        # Spectrogram extractor 
        self.spectrogram_extractor = Spectrogram(n_fft=self.window_size, hop_length=self.hop_size, 
            win_length=self.window_size, window=self.window, center=self.center, pad_mode=self.pad_mode, 
            freeze_parameters=True)

        # Logmel feature extractor
        self.logmel_extractor = LogmelFilterBank(sr=self.sample_rate, n_fft=self.window_size, 
            n_mels=self.mel_bins, fmin=self.fmin, fmax=self.fmax, ref=self.ref, amin=self.amin, 
            top_db=self.top_db, freeze_parameters=True)

        self.backbone =  backbone 
   
        self.sed_block = SED_Block(num_classes=num_classes, in_features=in_features, hidden_size=hidden_size, 
                activation = activation, drop_rate = drop_rate,   apply_attention=apply_attention)
                
    def call(self, input, mixup_fn=None, return_dict=False, training=False):
        """Input: (batch_size, length, data_length)
        """
         
        x = self.spectrogram_extractor(input)   # (batch_size, 1, time_steps, freq_bins)
        #print("sepctrogram :", x.shape)
        if self.use_logmel:
            x = self.logmel_extractor(x) # (batch_size, 1, time_steps, mel_bins)


        # (BS,  H, W, C)
        frames_num = x.shape[1]   
        x = tf.transpose(x, perm=[0, 1, 3, 2])
        x = self.bn(x) # BN applied on melbins
        x = tf.transpose(x, perm=[0, 1, 3, 2])
        
        if training and self.spectrogram_augmentation:
            x = self.spectrogram_augmentation(x)
        
        # Mixup on spectrogram
        if training and mixup_fn is not None:
            x = mixup_fn(x)  

        if x.shape[3] == 1:        
            x = tf.concat([x,x,x], axis=3)
        
        x = self.backbone(x) # (batch size, steps, freq, channels)
        
        outputs = self.sed_block(x, return_dict=return_dict)

        return outputs


class AudioClassifierSequenceSED(tf.keras.Model):
    def __init__(self, backbone):
        """Classifier for a new task using pretrained Cnn14 as a sub module.
        """
        super(AudioClassifierSequenceSED, self).__init__()        
        self.audio_backbone = backbone
        self.global_attention = [GlobalAttention(n) for n in backbone.num_classes]
        self.num_output = len(backbone.num_classes)

    def init_output(self, return_dict):
        outs = [] 
        if return_dict:
            for _ in range(self.num_output):
                outs.append({})
        else:
            for _ in range(self.num_output):
                outs.append([])
        return outs

    def merge_output(self, features, return_dict):
        # features: list of dict {"clipwise":clipwise_output, "segmentwise":segmentwise_output, "norm_att":norm_att}
        outs = self.init_output(return_dict)

        # iterate throught sequences
        if return_dict:
            for feat in features: # seq
                for i, feat_class in enumerate(feat): # class
                        for k, v in feat_class.items():
                            if k not in outs[i]:
                                outs[i][k] = []
                            outs[i][k].append(v)
            
            for i, o in enumerate(outs): # class
                for k in o.keys():
                    outs[i][k] = tf.stack(outs[i][k], axis=1)

        else:
            for feat in features: # seq
                for i, feat_class in enumerate(feat): # class
                    outs[i].append(feat_class)
            
            for i, o in enumerate(outs):
                    outs[i] = tf.stack(o, axis=1)
        return outs

    def call(self, input, mixup_fn=None, return_dict=False, training=False):
        """Input: (batch_size, length)
        """
        

        features = []
        for i in range(input.shape[1]):
            o = self.audio_backbone(input[:, i] , mixup_fn=mixup_fn, return_dict=return_dict,
                                     training=training)
            features.append(o)

        features = self.merge_output(features, return_dict=return_dict)
        
        for i, global_attn in enumerate(self.global_attention):
            if return_dict:
                features[i]["output"], features[i]["global_attention"] = global_attn(features[i]["clipwise"])
            else:
                features[i] = global_attn(features[i])[0]

        return features